# Copyright (C) 2001-2003 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# $Id: name.py,v 1.7 2003/05/30 08:44:50 halley Exp $

"""DNS Domain names."""

import string
import struct
import sys
import types

NAMERELN_NONE = 0
NAMERELN_SUPERDOMAIN = 1
NAMERELN_SUBDOMAIN = 2
NAMERELN_EQUAL = 3
NAMERELN_COMMONANCESTOR = 4

class EmptyLabel(Exception):
    """Raised if a label is empty."""
    pass

class BadEscape(Exception):
    """Raised if an escaped code in a text format name is invalid."""
    pass

class BadLabelType(Exception):
    """Raised if the label type of a wire format name is unknown."""
    pass

class NeedAbsoluteNameOrOrigin(Exception):
    """Raised if an attempt is made to convert a non-absolute name to
    wire when there is also a non-absolute (or missing) origin."""
    pass

class NameTooLong(Exception):
    """Raised if a name is > 255 octets long."""
    pass

class LabelTooLong(Exception):
    """Raised if a label is > 63 octets long."""
    pass

class MeaninglessComparison(Exception):
    """Raised if an absolute name is compared with a relative name."""
    pass

__escaped = {
    '"' : True,
    '(' : True,
    ')' : True,
    '.' : True,
    ';' : True,
    '\\' : True,
    '@' : True,
    '$' : True }

def _escapify(label):
    """Escape the characters in label which need it."""
    text = ''
    for c in label:
        if c in __escaped:
            text += '\\' + c
        elif ord(c) > 0x20 and ord(c) < 0x7F:
            text += c
        else:
            text += '\\%03d' % ord(c)
    return text

def validate_labels(labels):
    """Check for empty labels in the middle of a label sequence,
    labels that are too long, and for too many labels."""
    l = len(labels)
    if l > 255:
        raise NameTooLong
    i = -1
    j = 0
    for label in labels:
        if len(label) > 63:
            raise LabelTooLong
        if i < 0 and label == '':
            i = j
        j += 1
    if i >= 0 and i != l - 1:
        raise EmptyLabel

class Name(object):
    """A DNS domain name."""
    
    def __init__(self, labels):
        """Initialize a domain name from a list of labels."""
        validate_labels(labels)
        self.labels = labels[:]

    def is_absolute(self):
        """Is the most significant label of this name the root label?"""
        return len(self.labels) > 0 and self.labels[-1] == ''

    def is_wild(self):
        """Is this name wild?  (I.e. Is the least significant label '*'?)"""
        return self.labels[0] == '*'

    def __hash__(self):
        h = 0L
        for label in self.labels:
            for c in label:
                h += ( h << 3 ) + ord(c.lower())
        return int(h % sys.maxint)

    def is_comparable(self, other):
        """True iff. both names are absolute or both names are relative."""
        return self.is_absolute() == other.is_absolute()

    def fullcompare(self, other):
        """Compare two names, returning a 3-tuple (relation, order, nlabels).
        'relation' describes the relation ship beween the names, and is one of:
        NAMERELN_NONE, NAMERELN_SUPERDOMAIN, NAMERELN_SUBDOMAIN,
        NAMERELN_EQUAL, or NAMERELN_COMMONANCESTOR.  'order' is < 0 if
        self < other, > 0 if self > other, and == 0 if self == other.
        'nlabels' gives the number of significant labels that the two names
        have in common.  A relative name is always less than an absolute
        name"""

        sabs = self.is_absolute()
        oabs = other.is_absolute()
        if sabs != oabs:
            if sabs:
                return (NAMERELN_NONE, 1, 0)
            else:
                return (NAMERELN_NONE, -1, 0)
        l1 = len(self.labels)
        l2 = len(other.labels)
        ldiff = l1 - l2
        if ldiff < 0:
            l = l1
        else:
            l = l2

        order = 0
        nlabels = 0
        namereln = NAMERELN_NONE
        while l > 0:
            l -= 1
            l1 -= 1
            l2 -= 1
            label1 = self.labels[l1].lower()
            label2 = other.labels[l2].lower()
            if label1 < label2:
                order = -1
                if nlabels > 0:
                    namereln = NAMERELN_COMMONANCESTOR
                return (namereln, order, nlabels)
            elif label1 > label2:
                order = 1
                if nlabels > 0:
                    namereln = NAMERELN_COMMONANCESTOR
                return (namereln, order, nlabels)
            nlabels += 1
        if order == 0:
            order = ldiff
        if ldiff < 0:
            namereln = NAMERELN_SUPERDOMAIN
        elif ldiff > 0:
            namereln = NAMERELN_SUBDOMAIN
        else:
            namereln = NAMERELN_EQUAL
        return (namereln, order, nlabels)

    def is_subdomain(self, other):
        (nr, o, nl) = self.fullcompare(other)
        if nr == NAMERELN_SUBDOMAIN or nr == NAMERELN_EQUAL:
            return True
        return False

    def is_superdomain(self, other):
        (nr, o, nl) = self.fullcompare(other)
        if nr == NAMERELN_SUPERDOMAIN or nr == NAMERELN_EQUAL:
            return True
        return False

    def canonicalize(self):
        """Return a name which is equal to the current name, but is in
        canonical form."""
        return Name([x.lower() for x in self.labels])

    def __cmp__(self, other):
        return self.fullcompare(other)[1]
    
    def __repr__(self):
        return '<DNS name ' + self.__str__() + '>'
    
    def __str__(self):
        return self.to_text(False)

    def to_text(self, omit_final_dot = False):
        """Convert name to text format."""
        if len(self.labels) == 0:
            return '@'
        if len(self.labels) == 1 and self.labels[0] == '':
            return '.'
        if omit_final_dot and self.is_absolute():
            l = self.labels[:-1]
        else:
            l = self.labels
        s = string.join(map(_escapify, l), '.')
        return s

    def to_digestable(self):
        """Convert name to a format suitable for digesting in hashes."""
        labels = ["%s%s" % (chr(len(x)), x.lower()) for x in self.labels]
        return ''.join(labels)
        
    def to_wire(self, file, compress = None, origin = None):
        """Convert name to wire format, possibly compressing it.  If the
        name is relative and origin is not None, then origin will be appended
        to it."""
        if not self.is_absolute():
            if origin is None or not origin.is_absolute():
                raise NeedAbsoluteNameOrOrigin
            labels = self.labels[:]
            labels.extend(origin)
        else:
            labels = self.labels
        i = 0
        for label in labels:
            n = Name(labels[i:])
            i += 1
            if not compress is None:
                pos = compress.get(n)
            else:
                pos = None
            if not pos is None:
                value = 0xc000 + pos
                s = struct.pack('!H', value)
                file.write(s)
                return
            else:
                if not compress is None and len(n) > 1:
                    pos = file.tell()
                    if pos < 0xc000:
                        compress[n] = pos
                l = len(label)
                file.write(chr(l))
                if l > 0:
                    file.write(label)

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, index):
        return self.labels[index]

    def __getslice__(self, start, stop):
        return self.labels[start:stop]

    def __setitem__(self, index, value):
        if isinstance(index, types.SliceType):
            labels = self.labels[:]
            labels[index.start:index.stop] = value
            validate_labels(labels)
            self.labels = labels
        else:
            self.labels[index] = value
        
    def __add__(self, other):
        return self.concatenate(other)

    def __sub__(self, other):
        return self.relativize(other)

    def split(self, depth):
        "Split a name into a prefix and suffix at depth."""
        return (Name(self[: -depth]), Name(self[-depth :]))

    def concatenate(self, other):
        """Return a new name which is the concatenation of self and other."""
        if self.is_absolute():
            raise BadAbsoluteSliceAssignment
        n = Name(self.labels)
        n[len(n):] = list(other.labels)
        return n

    def relativize(self, origin):
        """If self is a subdomain of origin, return a new name which is self
        relative to origin.  Otherwise return a copy of self."""
        if not origin is None and self.is_subdomain(origin):
            return Name(self[: -len(origin)])
        else:
            return Name(self.labels)

    def derelativize(self, origin):
        """If self is a relative name, return a new name which is the
        concatenation of self and origin.  Otherwise return a copy of self."""
        if not self.is_absolute():
            return self.concatenate(origin)
        else:
            return Name(self.labels)

    def choose_relativity(self, origin=None, relativize=True):
        """Return a name with the relativity desired by the caller.  If
        origin is None, then a copy of self is returned.  Otherwise, if
        relativize is true the name is relativized, and if relativize is
        false the name is derelativized."""
        if origin:
            if relativize:
                return self.relativize(origin)
            else:
                return self.derelativize(origin)
        else:
            return Name(self.labels)

class ImmutableName(Name):
    """A DNS name which cannot be changed."""
    
    def __init__(self, labels):
        Name.__setattr__(self, 'labels', tuple(labels))

    def __setattr__(self, name, value):
        raise TypeError, "object doesn't support attribute assignment"

root = ImmutableName([''])
empty = ImmutableName([])

def from_text(text, origin = root):
    """Convert text into a Name object."""
    assert isinstance(text, str)
    assert origin is None or isinstance(origin, Name)
    labels = []
    label = ''
    escaping = False
    if text:
        if text == '.':
            return Name([''])
        for c in text:
            if escaping:
                if edigits == 0:
                    if c.isdigit():
                        total = int(c)
                        edigits += 1
                    else:
                        label += c
                        escaping = False
                else:
                    if not c.isdigit():
                        raise BadEscape
                    total *= 10
                    total += int(c)
                    edigits += 1
                    if edigits == 3:
                        escaping = False
                        label += chr(total)
            elif c == '.':
                if len(label) == 0:
                    raise EmptyLabel
                labels.append(label)
                label = ''
            elif c == '\\':
                escaping = True
                edigits = 0
                total = 0
            else:
                label += c
        if len(label) > 0:
            labels.append(label)
        else:
            labels.append('')
    if (len(labels) == 0 or labels[-1] != '') and not origin is None:
        labels.extend(origin.labels)
    return Name(labels)

def from_wire(message, current):
    """Convert possibly compressed wire format into a Name."""
    assert isinstance(message, str)
    labels = []
    biggest_pointer = current
    hops = 0
    count = ord(message[current])
    current += 1
    cused = 1
    while count != 0:
        if count < 64:
            labels.append(message[current : current + count])
            current += count
            if hops == 0:
                cused += count
        elif count >= 192:
            current = (count & 0x3f) * 256 + ord(message[current])
            if hops == 0:
                cused += 1
            if current >= biggest_pointer:
                raise BadPointer
            biggest_pointer = current
            hops += 1
        else:
            raise BadLabelType
        count = ord(message[current])
        current += 1
        if hops == 0:
            cused += 1
    labels.append('')
    return (Name(labels), cused)
